/*
 * Copyright (C) 2009 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
#include <string.h>
#include <jni.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <linux/fcntl.h>
#include <stdlib.h>
#include <time.h>
#include <linux/socket.h>
#include <sys/socket.h>
#include <pthread.h>
#include <sys/poll.h>
#include <poll.h>
#include <utils/Log.h>
#include <errno.h>
#include "mtk_log_ext.h"
#include "mtvdo.h"
#include "mtvdecex.h"
#include "mtvss.h"

#ifndef NELEM
#define NELEM(x) ((int) (sizeof(x) / sizeof((x)[0])))
#endif

#define RGB_BIT_PER_PIXEL 24
#define BGRA_BIT_PER_PIXEL 32
#define ROUNDUPFIXSIZE 4

typedef unsigned short WORD;
typedef unsigned int DWORD;
typedef int LONG;

static void *read_file(const char *fn, unsigned *_sz);
static void nativeCapturePictureOnce(JNIEnv* env, jobject thiz, int msi_on,
		int top, int left, int width, int height);
static void nativeCaptureScreen(JNIEnv* env, jobject thiz, int magic, int top,
		int left, int width, int height);
static int native_open(JNIEnv* env, jobject thiz, jint msi_on);
static void native_close(JNIEnv* env, jobject thiz, jint msi_on);

static JNINativeMethod gMethods[] = {
		{ "nativeCaptureScreenOnce", "(IIIII)V", (void*) nativeCapturePictureOnce },
		{ "nativeCaptureScreen", "(IIIII)V", (void*) nativeCaptureScreen },
		{ "nativeOpen", "(I)V", (void*) native_open },
		{ "nativeClose", "(I)V", (void*) native_close }
};

struct fields_t {
	jmethodID post_event;
};

typedef struct tagBITMAPFILEHEADER {
	WORD bfType;
	DWORD bfSize;
	WORD bfReserved1;
	WORD bfReserved2;
	DWORD bfOffBytes;
} BITMAPFILEHEADER, *PBITMAPFILEHEADER;

typedef struct tagBITMAPINFOHEADER {
	DWORD biSize;
	LONG biWidth;
	LONG biHeight;
	WORD biPlanes;
	WORD biBitCount;
	DWORD bitCompression;
	DWORD bitSizeImage;
	LONG biXPelsPerMeter;
	LONG bitYPelsPerMeter;
	DWORD biClrUsed;
	DWORD biClrImportant;
} BITMAPINFOHEADER, *PBITMAPINFOHEADER;

typedef struct tagRGBABITMAPINFOHEADER {
	BITMAPINFOHEADER base;
	DWORD redMask;
	DWORD greenMask;
	DWORD blueMask;
	DWORD alphaMask;
	DWORD LCS_WINDOWS_COLOR_SPACE;
	DWORD CIEXYZTRIPLE[9];
	DWORD redGamma;
	DWORD greenGamma;
	DWORD blueGamma;
} RGBABITMAPINFOHEADER, *PRGBABITMAPINFOHEADER;

enum {
	BI_RGB = 0,
	BI_RLE8,
	BI_RLE4,
	BI_BITFIELDS,
	BI_JPEG,
	BI_PNG,
	BI_ALPHABITFIELDS
};

static struct fields_t fields;
static int signal_fd = -1;
static int signal_recv_fd = -1;
static struct pollfd ufds[1];
static pthread_mutex_t lock;
static JavaVM* g_vm;
static jweak g_weak_cs = NULL;
//static jobject g_cs = NULL;
static int thread_running = 0;
static int g_msi_on;
static BOOL bmustong = FALSE;
static MTVSS_PATH_T capture = MTVSS_MON_OUT;

static BITMAPFILEHEADER bmp_file_header = {
		bfType:0x4D42,
		bfReserved1:0,
		bfReserved2:0,
		bfOffBytes:0x36
};

static BITMAPINFOHEADER bmp_info_header = {
		biSize:0x28,
		biPlanes:1,
		biBitCount:0x20,
		bitCompression:BI_RGB,
		biXPelsPerMeter:0,
		bitYPelsPerMeter:0,
		biClrUsed:0,
		biClrImportant:0
};

static BITMAPFILEHEADER rgba_bmp_file_header = {
		bfType:0x4D42,
		bfReserved1:0,
		bfReserved2:0,
		bfOffBytes:122
};

static RGBABITMAPINFOHEADER rgba_bmp_info_header = {
		base: {
			biSize:108,
			biPlanes:1,
			biBitCount:32,
			bitCompression:BI_RGB,
			biXPelsPerMeter:0,
			bitYPelsPerMeter:0,
			biClrUsed:0,
			biClrImportant:0
		},
		redMask:0xff000000,
		greenMask:0xff0000,
		blueMask:0xff00,
		alphaMask:0xff,
		LCS_WINDOWS_COLOR_SPACE:0,
		CIEXYZTRIPLE: { 0, },
		redGamma:0,
		greenGamma:0,
		blueGamma:0
};

static void* create_bitmap_rgb_header(int width, int height, int *_sz) {
	unsigned char *header = NULL;
	int row_size = (width * (BGRA_BIT_PER_PIXEL >> 3) + (ROUNDUPFIXSIZE - 1))
			& ~(ROUNDUPFIXSIZE - 1);
	bmp_info_header.biWidth = width;
	bmp_info_header.biHeight = height*-1;
	bmp_info_header.bitSizeImage = row_size * height;
	bmp_file_header.bfSize = bmp_info_header.bitSizeImage
			+ bmp_file_header.bfOffBytes;

	header = (unsigned char*) malloc(bmp_file_header.bfOffBytes);
	if (NULL != header) {
		unsigned char *p = header;
		*(p++) = bmp_file_header.bfType & 0xff;
		*(p++) = bmp_file_header.bfType >> 8;
		void *src = (void*) ((int) &bmp_file_header + 4);
		bcopy(src, p, 12);
		p += 12;
		src = (void*) &bmp_info_header;
		bcopy(src, p, sizeof(BITMAPINFOHEADER));
		if (_sz)
			*_sz = bmp_file_header.bfOffBytes;
	} else {
		if (_sz)
			*_sz = 0;
	}
	return (void*) header;
}

static void* create_bitmap_rgba_header(int width, int height, int *_sz) {
	unsigned char *header = NULL;
	int row_size = (width * (BGRA_BIT_PER_PIXEL >> 3) + (ROUNDUPFIXSIZE - 1))
			& ~(ROUNDUPFIXSIZE - 1);
	rgba_bmp_info_header.base.biWidth = width;
	rgba_bmp_info_header.base.biHeight = height*-1;
	rgba_bmp_info_header.base.bitSizeImage = row_size * height;
	rgba_bmp_file_header.bfSize = rgba_bmp_info_header.base.bitSizeImage
			+ rgba_bmp_file_header.bfOffBytes;

	header = (unsigned char*) malloc(rgba_bmp_file_header.bfOffBytes);
	if (NULL != header) {
		unsigned char *p = header;
		*(p++) = rgba_bmp_file_header.bfType & 0xff;
		*(p++) = rgba_bmp_file_header.bfType >> 8;
		void *src = (void*) ((int) &rgba_bmp_file_header + 4);
		bcopy(src, p, sizeof(BITMAPFILEHEADER) - 4);
		p += 12;
		src = (void*) &rgba_bmp_info_header;
		bcopy(src, p, sizeof(RGBABITMAPINFOHEADER));
		if (_sz)
			*_sz = rgba_bmp_file_header.bfOffBytes;
	} else {
		if (_sz)
			*_sz = 0;
	}
	return (void*) header;
}

#define JPG_PATH "/mnt/usbdisk/bitmap.bmp"

static int init_mtvss(int msi) {
    int ret=0;
    
    MTVSS_Init();
    LOGD("init_mtvss");	

    ret = MTVSS_SetColorFormat(0x00020001);            
	LOGD("MTVSS_SetColorFormat ret=%d",ret);
	return ret;
}

static int connect_mtvss(int msi) {
    int ret=0;
    LOGD("connect_mtvss");	

	//set raw data format to BGRAs
	if(TRUE == bmustong){
		ret = MTVSS_ConnectAll(capture);
		LOGD("MTVSS_ConnectAll ret=%d",ret);	
	}else{				
        ret = MTVSS_Connect(MTVDO_SUB);
        LOGD("MTVSS_Connect ret=%d",ret);	
    }

	return ret;
}

static void uninit_mtvss(int msi) {
	LOGD("uninit_mtvss");
    int ret =0;
    if(TRUE == bmustong){
        ret = MTVSS_DisconnectAll(capture);
        LOGD("MTVSS_DisconnectAll capture=%d, ret=%d",capture,ret);
    }
    else
    {	
        ret = MTVSS_Disconnect(MTVDO_SUB);
        LOGD("MTVSS_Disconnect ret=%d",ret);
        
    	if (!msi) {
    		MTVDO_REGION_T rRegionOut;
    		rRegionOut.u4X = 0; 
    		rRegionOut.u4Y = 0; 
    		rRegionOut.u4Width = 0; 
    		rRegionOut.u4Height = 0;
    
			ret = MTVDO_SetOutRegion(MTVDO_SUB, &rRegionOut);        
			LOGD("MTVDO_SetOutRegion ret=%d",ret);  

			ret= MTVDO_SetDisplayOff_Ext(MTVDO_SUB,0x02,FALSE);
    		LOGD("MTVDO_SetDisplayOff_Ext ret=%d", ret);
			
			ret = MTVDECEX_SetVideoSrc(MTVDO_SUB, MTVDECEX_SRC_MAX);        
			LOGD("MTVDECEX_SetVideoSrc ret=%d",ret);   
    	}
    }
}

#define QUERY_TIME_STEP 30000L

static void nativeCapturePictureOnce(JNIEnv* env, jobject thiz, int msi_on,
		int top, int left, int width, int height) {

	MTVDO_REGION_T rRegionOut;
	MTVDO_REGION_T rRegionSrcr;
	UINT16 w =0, h =0, color =0, source =0;
	UINT8 index =0, index_2 =0;
	UINT32 addrBase =0, addrBase_2 =0;
	UINT8 *raw = NULL;
	jbyteArray obj = NULL;
	jbyte* header = NULL;
	jbyte* data = NULL;
	int size = 0;
	int sz = 0;
	int mirror = 0;
	int flip = 0;
	int raw_width =0, raw_height =0;
	long timeout = 0;
	int ret =0;
    MTVSS_SCREEN_CAPTURE_MODE_T	eCaptureMode;
    MTVDO_REGION_T rgn;
	rgn.u4X = 0;
	rgn.u4Y = 0;
	rgn.u4Width = 1920;
	rgn.u4Height = 1080;
    
    /* drv init must do befort MTVSS_GetScreenCaptureMode*/
	DRV_Init();

    ret = MTVSS_GetScreenCaptureMode(&eCaptureMode);		
    LOGD("eCaptureMode = %d ret=%d", eCaptureMode, ret);
    if(eCaptureMode == MTVSS_SCREEN_CAPTURE_FROM_OSTG){
        bmustong = TRUE;
    }else{
        bmustong = FALSE;
    }

	if(init_mtvss(msi_on) < 0){
		return;
	}

    if(connect_mtvss(msi_on) < 0){
		return;
	}
	
    if(TRUE == bmustong){
        if (!msi_on){
            ret = MTVDO_SetMonOutEnable(1);
            LOGD("MTVDO_SetMonOutEnable: ret=%d", ret);
            
            ret = MTVDO_SetMonOutSource(7);
            LOGD("MTVDO_SetMonOutSource: ret=%d", ret);

            ret = MTVDO_SetMonOutOutputWindow(&rgn);
            LOGD("MTVDO_SetMonOutOutputWindow: ret=%d", ret);      	
        }
    }else{
    	if (!msi_on) {
    		//Bind OSTG with SUB
    		MTVDECEX_SetVideoSrc(MTVDO_SUB, MTVDECEX_SRC_OSDENCODE1);

			ret= MTVDO_SetDisplayOff_Ext(MTVDO_SUB,0x02,TRUE);
    		LOGD("MTVDO_SetDisplayOff_Ext ret=%d", ret);
			
    		ret = MTVDO_GetPlaneWH(&raw_width, &raw_height);
    		LOGD("Panel width:%d, height:%d ret=%d", raw_width, raw_height, ret);
    		rRegionOut.u4X = top * 10000 / raw_width;
    		rRegionOut.u4Y = left * 10000 / raw_height;
    		rRegionOut.u4Width = width * 10000 / raw_width;
    		rRegionOut.u4Height = height * 10000 / raw_height;

    		rRegionSrcr.u4X = 0;
    		rRegionSrcr.u4Y = 0;
    		rRegionSrcr.u4Width = 10000;
    		rRegionSrcr.u4Height = 10000;

    		LOGD("Set out Region, width:%d, height:%d", rRegionOut.u4Width,
    				rRegionOut.u4Height);

            ret = MTVDO_SetOutRegion(MTVDO_SUB, &rRegionOut);		
            LOGD("MTVDO_SetOutRegion: ret=%d", ret);

            ret = MTVDO_SetSrcRegion(MTVDO_SUB, &rRegionSrcr);		
            LOGD("MTVDO_SetSrcRegion ret=%d", ret);
        }
	}

	usleep(80*1000);/*driver may not init complete*/
	if(connect_mtvss(msi_on) < 0){
		return;
	}

	ret = MTVSS_GetFrameBufferIndex(&index);		
	LOGD("MTVSS_GetFrameBufferIndex: ret=%d", ret);
	ret = MTVSS_SetTrigger();
	LOGD("MTVSS_SetTrigger: ret=%d", ret);
	
	do {
		if (timeout >= QUERY_TIME_STEP * 20) {
			LOGD("Investigation time out.");
			goto oops;
		}
		usleep(QUERY_TIME_STEP);
		timeout += QUERY_TIME_STEP;
		ret = MTVSS_GetFrameBufferIndex(&index_2);
		LOGD("Data not update, investigating index=%d, index_2=%d...ret=%d", index, index_2, ret);
	} while (index == index_2);

	ret = MTVSS_GetVideoInfo(&w, &h, &color, &source);
	LOGD("MTVSS_GetVideoInfo w:%d, h:%d, color format:0x%x, source:%d, ret=%d", w, h,
			color, source, ret);

    if(TRUE == bmustong){        
        raw = (UINT8 *)malloc(1920*1080*4*sizeof(UINT8));
        if(raw == NULL)
        {
            LOGD("no buffer for raw.");
        }

        ret = MTVSS_GetFrameBuffer(0, 0, w, h, raw);
        LOGD("MTVSS_GetFrameBuffer : width %d, height %d ret = %d", w, h, ret);
        
        LOGD("Dump successful!!");


#if 0
FILE *vssdump2 = NULL;

        vssdump2 = fopen("111.raw", "w+");
        
        if(vssdump2 == NULL)
      	{
      		printf("failed to open file %s!\n", "000.raw");
      		return;
      	}        
        
        fwrite(raw, sizeof(UINT8), 4*w*h, vssdump2);
        fflush(vssdump2);

        fclose(vssdump2);
#endif
		
    }
    else
    {	
	    raw = (UINT8 *)malloc(w*h*4*sizeof(UINT8));
		if(raw == NULL)
		{
			LOGD("no buffer for raw.w=%d, h=%d", w, h);
		}

		ret = MTVSS_GetFrameBuffer(0, 0, w, h, raw);
		LOGD("MTVSS_GetFrameBuffer : width %d, height %d ret = %d", w, h, ret);

		LOGD("Dump successful!!");
    	// WTF...
    	/*
	    	{
	    		w = 1280;
	    		h = 720;
	    	}
	    	ret = MTVDO_GetMirrorFlip(ENUM_FLIP_TYPE, &flip);			
	    	LOGD("MTVDO_GetMirrorFlip.ret=%d", ret);
	        
	    	ret = MTVDO_GetMirrorFlip(ENUM_MIRROR_TYPE, &mirror);
	    	LOGD("MTVDO_GetMirrorFlip time out.ret=%d", ret);

	    	LOGD("flip:%d, mirror:%d", flip, mirror);
	        
	    	ret = MTVSS_GetAddrBase(&addrBase, &addrBase_2);	
	    	LOGD("MTVSS_GetAddrBase.ret=%d", ret);

	    	ret = MTVSS_Phy2Virt(addrBase, (UINT32 *)raw);	
	    	LOGD("MTVSS_Phy2Virt.ret=%d", ret);

	        if(index_2 == 1) {
	    		raw += addrBase_2 - addrBase;
	    	}
	    	*/
    }

	header = (jbyte*) create_bitmap_rgb_header(w, h, &sz);
	if (header) {
		data = (jbyte*) raw;
		if (data) {
			size = w * h * (BGRA_BIT_PER_PIXEL >> 3);
			obj = (*env)->NewByteArray(env, size + sz);
			if (obj) {
				jbyte* buffer = (jbyte*) malloc(size + sz);
				if (buffer) {
//					bcopy(header, buffer, sz);
//					bcopy(data, buffer + sz, size);
					memcpy(buffer, header, sz);
					memcpy(buffer + sz, data, size);
					(*env)->SetByteArrayRegion(env, obj, 0, size + sz, buffer);
					(*env)->CallVoidMethod(env, thiz, fields.post_event, 0, flip,
							mirror, obj);
					free(buffer);
				}
			}
		}
	}
oops:
	uninit_mtvss(msi_on);
	if (obj) {
		(*env)->DeleteLocalRef(env, obj);
	}
	if (header){
		free(header);
    }
    if(raw){
        free(raw);
    }
}

static int handle_request(JNIEnv* env, int msi_on, int magic, int top, int left,
		int width, int height) {
	MTVDO_REGION_T rRegionOut;
	MTVDO_REGION_T rRegionSrcr;
	UINT16 w, h, color, source;
	UINT8 index, index_2;
	UINT32 addrBase =0, addrBase_2 =0;
	UINT8 *raw = NULL;
	jbyteArray obj = NULL;
	jbyte* header = NULL;
	jbyte* data = NULL;
	int size = 0;
	int sz = 0;
	int mirror = 0;
	int flip = 0;
	int ret = 0;
	long timeout = 0;
	int raw_width =0, raw_height =0;
    MTVDO_REGION_T rgn;
	rgn.u4X = 0;
	rgn.u4Y = 0;
	rgn.u4Width = 1920;
	rgn.u4Height = 1080;
	LOGD("Handle capture request, magic:%d, width:%d, height:%d", magic, width, height);
	
    if(TRUE == bmustong){
        if (!msi_on){
            ret = MTVDO_SetMonOutEnable(1);
            LOGD("MTVDO_SetMonOutEnable: ret=%d", ret);
            
            ret = MTVDO_SetMonOutSource(7);
            LOGD("MTVDO_SetMonOutSource: ret=%d", ret);

            ret = MTVDO_SetMonOutOutputWindow(&rgn);
            LOGD("MTVDO_SetMonOutOutputWindow: ret=%d", ret);
        }
    }else{
    	if (!msi_on) {
    		//Bind OSTG with SUB
    		MTVDECEX_SetVideoSrc(MTVDO_SUB, MTVDECEX_SRC_OSDENCODE1);

			ret= MTVDO_SetDisplayOff_Ext(MTVDO_SUB,0x02,TRUE);
    		LOGD("MTVDO_SetDisplayOff_Ext ret=%d", ret);
			
    		ret = MTVDO_GetPlaneWH(&raw_width, &raw_height);
    		LOGD("Panel width:%d, height:%d ret=%d", raw_width, raw_height, ret);
    		rRegionOut.u4X = top * 10000 / raw_width;
    		rRegionOut.u4Y = left * 10000 / raw_height;
    		rRegionOut.u4Width = width * 10000 / raw_width;
    		rRegionOut.u4Height = height * 10000 / raw_height;

    		rRegionSrcr.u4X = 0;
    		rRegionSrcr.u4Y = 0;
    		rRegionSrcr.u4Width = 10000;
    		rRegionSrcr.u4Height = 10000;

    		LOGD("Set out Region, width:%d, height:%d", rRegionOut.u4Width,
    				rRegionOut.u4Height);

            ret = MTVDO_SetOutRegion(MTVDO_SUB, &rRegionOut);		
            LOGD("MTVDO_SetOutRegion: ret=%d", ret);

            ret = MTVDO_SetSrcRegion(MTVDO_SUB, &rRegionSrcr);		
            LOGD("MTVDO_SetSrcRegion ret=%d", ret);
        }
	}

	usleep(80*1000);/*driver may not init complete*/
	if(connect_mtvss(msi_on) < 0){
		return -1;
	}
	
	ret = MTVSS_GetFrameBufferIndex(&index);		
	LOGD("MTVSS_GetFrameBufferIndex ret=%d", ret);

	ret = MTVSS_SetTrigger();		
	LOGD("MTVSS_SetTrigger ret=%d", ret);

	do {
		if (timeout >= QUERY_TIME_STEP * 20) {
			LOGD("Investigation time out.");
			return -1;
		}
		usleep(QUERY_TIME_STEP);
		timeout += QUERY_TIME_STEP;
		ret = MTVSS_GetFrameBufferIndex(&index_2);
		LOGD("Data not update, investigating ...ret=%d", ret);
	} while (index == index_2);

	MTVSS_GetVideoInfo(&w, &h, &color, &source);
	LOGD("MTVSS_GetVideoInfo w:%d, h:%d, color format:0x%x, source:%d", w, h,
			color, source);

    if(TRUE == bmustong){
        raw = (UINT8 *)malloc(1920*1080*sizeof(UINT8));
        if(raw == NULL)
        {
            LOGD("no buffer for raw.");
        }

        ret = MTVSS_GetFrameBuffer(0, 0, w, h, raw);
        LOGD("MTVSS_GetFrameBuffer : width %d, height %d ret = %d", w, h, ret);
        
        LOGD("Dump successful!!");
    }
    else
    {	
        raw = (UINT8 *)malloc(1920*1080*sizeof(UINT8));
        if(raw == NULL)
        {
            LOGD("no buffer for raw.");
        }
        // WTF...
    	{
    		w = 1280;
    		h = 720;
    	}
    	MTVDO_GetMirrorFlip(ENUM_FLIP_TYPE, &flip);
    	MTVDO_GetMirrorFlip(ENUM_MIRROR_TYPE, &mirror);
    	LOGD("flip:%d, mirror:%d", flip, mirror);
    	MTVSS_GetAddrBase(&addrBase, &addrBase_2);
    	MTVSS_Phy2Virt(addrBase, (UINT32 *)raw);
    	if(index_2 == 1) {
    		raw += addrBase_2 - addrBase;
    	}
    }
	header = (jbyte*) create_bitmap_rgb_header(w, h, &sz);
	if (header) {
		data = (jbyte*) raw;
		if (data) {
			size = w * h * (BGRA_BIT_PER_PIXEL >> 3);
			obj = (*env)->NewByteArray(env, size + sz);
			if (obj) {
				jbyte* buffer = (jbyte*) malloc(size + sz);
				if (buffer) {
//					bcopy(header, buffer, sz);
//					bcopy(data, buffer + sz, size);
					memcpy(buffer, header, sz);
					memcpy(buffer + sz, data, size);
					(*env)->SetByteArrayRegion(env, obj, 0, size + sz, buffer);
					if (!(*env)->IsSameObject(env, g_weak_cs, NULL )) {
						jobject cs = (*env)->NewLocalRef(env, g_weak_cs);
						if (cs) {
							(*env)->CallVoidMethod(env, cs, fields.post_event,
									magic, flip, mirror, obj);
							(*env)->DeleteLocalRef(env, cs);
							return 1;
						}
					} else {
						LOGD("Global weak reference have been collected.");
					}
					free(buffer);
				}
			}
		}
	}
	if (obj) {
		(*env)->DeleteLocalRef(env, obj);
	}
    if(raw){
        free(raw);
    }
	return 0;
}

typedef void *(*android_thread_func_t)(void *);

typedef struct {
	int magic;
	int top;
	int left;
	int width;
	int height;
} request_mtvss_data_t, *request_mtvss_data_t_ptr;

static void signal_init(void) {
	int s[2];
	/* create a signalling mechanism for the sigchld handler */
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, s) == 0) {
		signal_fd = s[0];
		signal_recv_fd = s[1];
		fcntl(s[0], F_SETFD, FD_CLOEXEC);
		fcntl(s[0], F_SETFL, O_NONBLOCK);
		fcntl(s[1], F_SETFD, FD_CLOEXEC);
		fcntl(s[1], F_SETFL, O_NONBLOCK);
	}
}

static void* _thread_loop(void *arg) {
	int i;
	int res;
	JNIEnv *env;
	int msi_on = *(int*) arg;
	int ret;
    MTVSS_SCREEN_CAPTURE_MODE_T	eCaptureMode;
    
    /* drv init must do befort MTVSS_GetScreenCaptureMode*/
	DRV_Init();

    ret = MTVSS_GetScreenCaptureMode(&eCaptureMode);		
    LOGD("eCaptureMode = %d ret=%d", eCaptureMode, ret);
    if(eCaptureMode == MTVSS_SCREEN_CAPTURE_FROM_OSTG){
        bmustong = TRUE;
    }
    else
    {
        bmustong = FALSE;
    }
	
	if(init_mtvss(msi_on) < 0){
		return -1;
	}

    if(connect_mtvss(msi_on) < 0){
		return -1;
	}

	thread_running = 1;
	LOGD("Enter thread loop ...");
	// attach current native thread to VM.
	(*g_vm)->AttachCurrentThread(g_vm, &env, NULL );
	if (NULL == env) {
		goto oops;
	}

	if ((*env)->ExceptionOccurred(env)) {
		(*env)->ExceptionDescribe(env);
	}

	for (;;) {
		poll(ufds, NELEM(ufds), -1);
		for (i = 0; i < NELEM(ufds) ; i++) {
			if (ufds[i].revents == POLLIN) {
				if (ufds[i].fd == signal_recv_fd) {
					request_mtvss_data_t data;
					pthread_mutex_lock(&lock);
					read(signal_recv_fd, (void*) &data,
							sizeof(request_mtvss_data_t));
					pthread_mutex_unlock(&lock);
					if (data.magic < 0) {
						LOGD("Quit thread loop.");
						goto oops;
					} else {
						if (!handle_request(env, msi_on, data.magic, data.top,
								data.left, data.width, data.height)) {
							LOGD("handle_request error.");
							goto oops;
						}
					}
				}
			}
		}
	}
	oops: uninit_mtvss(msi_on);
	if (g_weak_cs)
		(*env)->DeleteWeakGlobalRef(env, g_weak_cs);
//	if (g_cs)
//		(*env)->DeleteGlobalRef(env, g_cs);
	(*g_vm)->DetachCurrentThread(g_vm);
	thread_running = 0;
	return 0;
}

static int createRawThreadEtc(android_thread_func_t entry_func, void *user_data,
		const char* thread_name, int32_t thread_priority,
		size_t thread_static_size, pthread_t *threadid) {
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);

	if (thread_static_size) {
		pthread_attr_setstacksize(&attr, thread_static_size);
	}

	errno = 0;
	pthread_t thread;
	int result = pthread_create(&thread, &attr,
			(android_thread_func_t) entry_func, user_data);
	pthread_attr_destroy(&attr);
	if (result != 0) {
		LOGE("createRawThreadEtc failed (entry=%p, res=%d, errno=%d)\n"
				"(android thread_priority=%d)", entry_func, result, errno,
				thread_priority);
		return 0;
	}
	if (threadid) {
		*threadid = thread;
	}
	return 1;
}

static int native_open(JNIEnv* env, jobject thiz, jint msi_on) {
	pthread_mutex_init(&lock, NULL );
	signal_init();
	ufds[0].fd = signal_recv_fd;
	ufds[0].events = POLLIN;
	ufds[0].revents = 0;

	g_msi_on = msi_on;
	g_weak_cs = (*env)->NewWeakGlobalRef(env, thiz);
//	g_cs = (*env)->NewGlobalRef(env, thiz);
	int result = createRawThreadEtc(_thread_loop, &g_msi_on,
			"capture_screen_thread", 0, 0, 0);
	return result;
}

static void native_close(JNIEnv* env, jobject thiz, jint msi_on) {
	request_mtvss_data_t data;
	data.magic = -1;
	pthread_mutex_lock(&lock);
	write(signal_fd, (void*) &data, sizeof(request_mtvss_data_t));
	pthread_mutex_unlock(&lock);
}

static void nativeCaptureScreen(JNIEnv* env, jobject thiz, int magic, int top,
		int left, int width, int height) {
	request_mtvss_data_t data;
	data.magic = magic;
	data.top = top;
	data.left = left;
	data.width = width;
	data.height = height;

	pthread_mutex_lock(&lock);
	write(signal_fd, (void*) &data, sizeof(request_mtvss_data_t));
	pthread_mutex_unlock(&lock);
}

static void *read_file(const char *fn, unsigned *_sz) {
	char *data;
	int sz;
	int fd;

	data = 0;
	fd = open(fn, O_RDONLY);
	if (fd < 0) {
		return 0;
	}
	sz = lseek(fd, 0, SEEK_END);
	if (sz < 0)
		goto oops;

	if (lseek(fd, 0, SEEK_SET) != 0)
		goto oops;

	data = (char*) malloc(sz);
	if (data == 0)
		goto oops;

	if (read(fd, data, sz) != sz)
		goto oops;
	close(fd);
	if (_sz)
		*_sz = sz;
	return data;

	oops: close(fd);
	if (data != 0)
		free(data);
	return 0;
}

#ifdef __cplusplus
extern "C" {
#endif

jint JNI_OnLoad(JavaVM* vm, void* reserved) {
	JNIEnv* env = NULL;
	g_vm = vm;
	if ((*vm)->GetEnv(vm, (void**) (&env), JNI_VERSION_1_6) != JNI_OK) {
		return -1;
	}
	jclass clazz = (*env)->FindClass(env, "com/mediatek/capture/CaptureScreen");
	if (NULL == clazz) {
		return -1;
	}

	(*env)->RegisterNatives(env, clazz, gMethods, NELEM(gMethods) );
	fields.post_event = (*env)->GetMethodID(env, clazz, "postEventFromNative",
			"(IIILjava/lang/Object;)V");

	if (NULL != clazz) {
		(*env)->DeleteLocalRef(env, clazz);
	}

	return JNI_VERSION_1_6;
}

void JNI_OnUnload(JavaVM* vm, void* reserved) {
	JNIEnv* env = NULL;
	if ((*vm)->GetEnv(vm, (void**) (&env), JNI_VERSION_1_6) != JNI_OK) {
		return;
	}
}

#ifdef __cplusplus
}
#endif
