/*
 * Copyright 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifdef LOG_TAG
#undef LOG_TAG
#endif
#define LOG_TAG "Lights"
#define DEBUG 0

#include <cutils/log.h>

#include <stdint.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>

#include <sys/ioctl.h>
#include <sys/types.h>

#include <hardware/lights.h>

#ifdef ANDROID_TWO_WORLDS
#include "u_common.h"
#include "app_util/mtktvapi/a_mtktvapi_highlevel.h"
#include "app_util/mtktvapi/a_mtktvapi_config.h"
#include "app_util/mtktvapi/mtktvapi_config_type.h"
#include "IMtkPb_Ctrl.h"
#include "c_svctx.h"
#include "c_handle.h"
#endif


/******************************************************************************/

static pthread_once_t g_init = PTHREAD_ONCE_INIT;
static pthread_mutex_t g_lock = PTHREAD_MUTEX_INITIALIZER;

/*
char const *const LCD_FILE           = "/sys/class/leds/lcd-backlight/brightness";
char const *const RED_LED_FILE       = "/sys/class/leds/red/brightness";
char const *const GREEN_LED_FILE     = "/sys/class/leds/green/brightness";
char const *const BLUE_LED_FILE      = "/sys/class/leds/blue/brightness";
char const *const RED_TIMEOUT_FILE   = "/sys/class/leds/red/on_off_ms";
char const *const GREEN_TIMEOUT_FILE = "/sys/class/leds/green/on_off_ms";
char const *const BLUE_TIMEOUT_FILE  = "/sys/class/leds/blue/on_off_ms";
char const *const RGB_LOCKED_FILE    = "/sys/class/leds/red/rgb_start";
*/

/**
 * device methods
 */

void init_globals(void)
{
    // init the mutex
    pthread_mutex_init(&g_lock, NULL);
    
    IMtkPb_Ctrl_Init();
}

/*
static int write_int(char const* path, int value)
{
    int fd;
    static int already_warned = 0;

    fd = open(path, O_RDWR);
    if (fd >= 0) {
        char buffer[32] = {0,};
        int bytes = snprintf(buffer, sizeof(buffer), "%d\n", value);
        int amt = write(fd, buffer, bytes);
        close(fd);
        return amt == -1 ? -errno : 0;
    } else {
        if (already_warned == 0) {
            ALOGE("write_int failed to open %s\n", path);
            already_warned = 1;
        }
        return -errno;
    }
}

static int write_on_off(char const* path, int on, int off)
{
    int fd;
    static int already_warned = 0;

    fd = open(path, O_RDWR);
    if (fd >= 0) {
        char buffer[32] = {0,};
        int bytes = snprintf(buffer, sizeof(buffer), "%d %d\n", on, off);
        int amt = write(fd, buffer, bytes);
        close(fd);
        return amt == -1 ? -errno : 0;
    } else {
        if (already_warned == 0) {
            ALOGE("write_int failed to open %s\n", path);
            already_warned = 1;
        }
        return -errno;
    }
}
*/

static int rgb_to_brightness(struct light_state_t const* state)
{
    int color = state->color & 0x00ffffff;

    return ((77 * ((color >> 16) & 0x00ff))
            + (150 * ((color >> 8) & 0x00ff)) + (29 * (color & 0x00ff))) >> 8;
}

static int set_light_backlight(struct light_device_t* dev,
        struct light_state_t const* state)
{
    int err = 0;
    int brightness = rgb_to_brightness(state);

    pthread_mutex_lock(&g_lock);
    #ifdef ANDROID_TWO_WORLDS
    ALOGD("%s: call a_mtktvapi_config_set_value(CFG_ANDROID_BACK_LIGHT, brightness = %d, CFGF_SET_VALUE | CFGF_UPDATE_VALUE).", __FUNCTION__, brightness);
    err = a_mtktvapi_config_set_value(-1, CFG_ANDROID_BACK_LIGHT, brightness, CFGF_SET_VALUE | CFGF_UPDATE_VALUE);
    #endif
    pthread_mutex_unlock(&g_lock);

    ALOGD("%s: return err = %d", __FUNCTION__, err);
    return err;
}

static int set_light_keyboard(struct light_device_t* dev,
        struct light_state_t const* state)
{
    /* @Waiting for later implementation. */
    ALOGD( "%s: Not implemented.", __FUNCTION__ );
    return 0;
}

static int set_light_buttons(struct light_device_t* dev,
        struct light_state_t const* state)
{
    /* @Waiting for later implementation. */
    ALOGD( "%s: Not implemented.", __FUNCTION__ );
    return 0;
}

static int set_light_battery(struct light_device_t* dev,
        struct light_state_t const* state)
{
    /* @Waiting for later implementation. */
    ALOGD( "%s: Not implemented.", __FUNCTION__ );
    return 0;
}

static int set_light_notifications(struct light_device_t* dev,
        struct light_state_t const* state)
{
    /* @Waiting for later implementation. */
    ALOGD( "%s: Not implemented.", __FUNCTION__ );
    return 0;
}

static int set_light_attention(struct light_device_t* dev,
        struct light_state_t const* state)
{
    /* @Waiting for later implementation. */
    ALOGD( "%s: Not implemented.", __FUNCTION__ );
    return 0;
}


/** Close the lights device */
static int close_lights(struct light_device_t *dev)
{
    if (dev)
        free(dev);

    return 0;
}

/******************************************************************************/

/**
 * module methods
 */

/** Open a new instance of a lights device using name */
static int open_lights(const struct hw_module_t* module, char const* name,
        struct hw_device_t** device)
{
    int (*set_light)(struct light_device_t* dev,
            struct light_state_t const* state);

    if (0 == strcmp( LIGHT_ID_BACKLIGHT, name )) {
        set_light = set_light_backlight;
    } else if (0 == strcmp( LIGHT_ID_KEYBOARD, name )) {
        set_light = set_light_keyboard;
    } else if (0 == strcmp( LIGHT_ID_BUTTONS, name )) {
        set_light = set_light_buttons;
    } else if (0 == strcmp( LIGHT_ID_BATTERY, name )) {
        set_light = set_light_battery;
    } else if (0 == strcmp( LIGHT_ID_NOTIFICATIONS, name )) {
        set_light = set_light_notifications;
    } else if (0 == strcmp( LIGHT_ID_ATTENTION, name )) {
        set_light = set_light_attention;
    } else {
        ALOGE( "%s: %s light isn't supported yet.", __FUNCTION__, name );
        return -EINVAL;
    }
    

    pthread_once(&g_init, init_globals);

    struct light_device_t *dev = malloc(sizeof(struct light_device_t));
    if(NULL == dev)
    {
        ALOGE("%s: malloc fail,return fail!", __FUNCTION__);
        return -1;
    }
    memset(dev, 0, sizeof(*dev));

    dev->common.tag = HARDWARE_DEVICE_TAG;
    dev->common.version = 0;
    dev->common.module = (struct hw_module_t*)module;
    dev->common.close = (int (*)(struct hw_device_t*))close_lights;
    dev->set_light = set_light;

    *device = (struct hw_device_t*)dev;

    return 0;
}

static struct hw_module_methods_t lights_module_methods = {
    .open =  open_lights,
};

/*
 * The lights Module
 */
struct hw_module_t HAL_MODULE_INFO_SYM = {
    .tag = HARDWARE_MODULE_TAG,
    .version_major = 1,
    .version_minor = 0,
    .id = LIGHTS_HARDWARE_MODULE_ID,
    .name = "lights Module",
    .author = "Google, Inc.",
    .methods = &lights_module_methods,
};
